/* Copyright (C) 2000 Red Hat, Inc.
 * Portions copyright (C) 2004 VMware, Inc.
 * This file is part of the Pixops Library.
 *
 * The pixops library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * The pixops library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * pixops.c --
 *
 *      Implements C routines for image scaling. Some functions also have an
 *      MMX-accelerated implementation, see the corresponding .S files for the
 *      code.
 */

#include <math.h>
#include <stdlib.h>

#include "pixops.h"
#include "pixops-internal.h"
#include "convert.h"

#define SUBSAMPLE_BITS 4
#define SUBSAMPLE (1 << SUBSAMPLE_BITS)
#define SUBSAMPLE_MASK ((1 << SUBSAMPLE_BITS)-1)
#define SCALE_SHIFT 16

static int
get_check_shift (int check_size)
{
  int check_shift = 0;
  return_val_if_fail (check_size >= 0, 4);

  while (!(check_size & 1))
    {
      check_shift++;
      check_size >>= 1;
    }

  return check_shift;
}

static void
pixops_scale_nearest (uint8 *dest_buf,
		      int render_x0,
		      int render_y0,
		      int render_x1,
		      int render_y1,
		      int dest_rowstride,
		      int dest_bytes_per_pixel,
                      int dest_depth,
		      boolean dest_has_alpha,
		      const uint8  *src_buf,
		      int src_width,
		      int src_height,
		      int src_rowstride,
		      int src_bytes_per_pixel,
                      int src_depth,
                      uint32 *src_colormap,
		      boolean src_has_alpha,
		      double scale_x,
		      double scale_y)
{
   int i, j;
   int x;
   int x_step = (1 << SCALE_SHIFT) / scale_x;
   int y_step = (1 << SCALE_SHIFT) / scale_y;
   int src_bppdepth = BPPDEPTH(src_bytes_per_pixel << 3, src_depth);
   int dest_bppdepth = BPPDEPTH(dest_bytes_per_pixel << 3, dest_depth);

   /*
    * Define this bogus type so the macro loop will work
    */
   typedef unsigned char uint24;

#define INNER_LOOP(op, SRC_BPP, DEST_BPP, INDEXED) \
  for (j=0; j < (render_x1 - render_x0); j++) { \
     const unsigned char *p = src + (x >> SCALE_SHIFT) * (SRC_BPP >> 3); \
     unsigned int pix; \
     if (SRC_BPP == 24 && INDEXED == FALSE) { \
        pix = op((p[0] | (p[1] << 8) | (p[2] << 16))); \
     } else if (INDEXED == TRUE) { \
        pix = op(src_colormap[*p]); \
     } else { \
        pix = op((*(uint##SRC_BPP *) p)); \
     } \
     if (DEST_BPP == 24) { \
        dest[0] = pix & 0xff; \
        dest[1] = (pix >> 8) & 0xff; \
        dest[2] = (pix >> 16) & 0xff; \
     } else { \
        *((uint##DEST_BPP *) dest) = (uint##DEST_BPP) pix; \
     } \
     dest += DEST_BPP >> 3; \
     x += x_step; \
  } \
  break;

   for (i = 0; i < (render_y1 - render_y0); i++) {

      const uint8 *src  = src_buf
         + (((i + render_y0) * y_step + y_step / 2) >> SCALE_SHIFT)
         * src_rowstride;

      uint8 *dest = dest_buf + i * dest_rowstride;

      x = render_x0 * x_step + x_step / 2;

      switch(src_bppdepth) {
      case 8:
	 switch(dest_bppdepth) {
	 case 8:
	    INNER_LOOP(CONVERT_NOP, 8, 8, TRUE);
	 case 15:
	    INNER_LOOP(CONVERT_24_TO_15, 8, 16, TRUE);
	 case 16:
	    INNER_LOOP(CONVERT_24_TO_16, 8, 16, TRUE);
	 case 24:
	    INNER_LOOP(CONVERT_NOP, 8, 24, TRUE);
	 case 32:
	    INNER_LOOP(CONVERT_NOP, 8, 32, TRUE);
	 }
	 break;
      case 15:
         switch(dest_bppdepth) {
	 case 8:
	    INNER_LOOP(CONVERT_15_TO_8BGR, 16, 8, FALSE);
         case 15:
            INNER_LOOP(CONVERT_NOP, 16, 16, FALSE);
         case 16:
            INNER_LOOP(CONVERT_15_TO_16, 16, 16, FALSE);
         case 24:
            INNER_LOOP(CONVERT_15_TO_24, 16, 24, FALSE);
         case 32:
            INNER_LOOP(CONVERT_15_TO_24, 16, 32, FALSE);
         }
         break;
      case 16:
	 switch(dest_bppdepth) {
	 case 8:
	    INNER_LOOP(CONVERT_16_TO_8BGR, 16, 8, FALSE);
         case 15:
            INNER_LOOP(CONVERT_16_TO_15, 16, 16, FALSE);
         case 16:
            INNER_LOOP(CONVERT_NOP, 16, 16, FALSE);
         case 24:
            INNER_LOOP(CONVERT_16_TO_24, 16, 24, FALSE);
         case 32:
            INNER_LOOP(CONVERT_16_TO_24, 16, 32, FALSE);
	 }
         break;
      case 24:
         switch(dest_bppdepth) {
	 case 8:
	    INNER_LOOP(CONVERT_24_TO_8BGR, 24, 8, FALSE);
         case 15:
            INNER_LOOP(CONVERT_24_TO_15, 24, 16, FALSE);
         case 16:
            INNER_LOOP(CONVERT_24_TO_16, 24, 16, FALSE);
         case 24:
            INNER_LOOP(CONVERT_NOP, 24, 24, FALSE);
         case 32:
            INNER_LOOP(CONVERT_NOP, 24, 32, FALSE);
         }
         break;
      case 32:
         switch(dest_bppdepth) {
	 case 8:
	    INNER_LOOP(CONVERT_24_TO_8BGR, 32, 8, FALSE);
         case 15:
            INNER_LOOP(CONVERT_24_TO_15, 32, 16, FALSE);
         case 16:
            INNER_LOOP(CONVERT_24_TO_16, 32, 16, FALSE);
         case 24:
            INNER_LOOP(CONVERT_NOP, 32, 24, FALSE);
         case 32:
            INNER_LOOP(CONVERT_NOP, 32, 32, FALSE);
         }
         break;
      }
   }
#undef INNER_LOOP
}


static void
pixops_scale_composite_nearest (uint8 *dest_buf,
                                int render_x0,
                                int render_y0,
                                int render_x1,
                                int render_y1,
                                int dest_rowstride,
                                int dest_channels,
                                boolean dest_has_alpha,
                                const uint8  *src_buf,
                                int src_width,
                                int src_height,
                                int src_rowstride,
                                int src_channels,
                                boolean src_has_alpha,
                                double scale_x,
                                double scale_y,
                                int overall_alpha)
{
  int i, j;
  int x;
  int x_step = (1 << SCALE_SHIFT) / scale_x;
  int y_step = (1 << SCALE_SHIFT) / scale_y;

  for (i = 0; i < (render_y1 - render_y0); i++) {

     const uint8 *src  = src_buf
        + (((i + render_y0) * y_step + y_step / 2) >> SCALE_SHIFT)
        * src_rowstride;

     uint8       *dest = dest_buf + i * dest_rowstride;

     x = render_x0 * x_step + x_step / 2;

     for (j=0; j < (render_x1 - render_x0); j++) {

        const uint8 *p = src + (x >> SCALE_SHIFT) * src_channels;
        unsigned int  a0;

        if (src_has_alpha)
           a0 = (p[3] * overall_alpha) / 0xff;
        else
           a0 = overall_alpha;

        switch (a0) {
        case 0:
           break;
        case 255:
           dest[0] = p[0];
           dest[1] = p[1];
           dest[2] = p[2];
           if (dest_has_alpha)
              dest[3] = 0xff;
           break;
        default:
           if (dest_has_alpha) {
              unsigned int w0 = 0xff * a0;
              unsigned int w1 = (0xff - a0) * dest[3];
              unsigned int w = w0 + w1;

              dest[0] = (w0 * p[0] + w1 * dest[0]) / w;
              dest[1] = (w0 * p[1] + w1 * dest[1]) / w;
              dest[2] = (w0 * p[2] + w1 * dest[2]) / w;
              dest[3] = w / 0xff;
           } else {
              unsigned int a1 = 0xff - a0;
              unsigned int tmp;

              tmp = a0 * p[0] + a1 * dest[0] + 0x80;
              dest[0] = (tmp + (tmp >> 8)) >> 8;
              tmp = a0 * p[1] + a1 * dest[1] + 0x80;
              dest[1] = (tmp + (tmp >> 8)) >> 8;
              tmp = a0 * p[2] + a1 * dest[2] + 0x80;
              dest[2] = (tmp + (tmp >> 8)) >> 8;
           }
           break;
        }
        dest += dest_channels;
        x += x_step;
     }
  }
}


static void
pixops_scale_composite_color_nearest (uint8 *dest_buf,
                                      int render_x0,
                                      int render_y0,
                                      int render_x1,
                                      int render_y1,
                                      int dest_rowstride,
                                      int dest_channels,
                                      boolean dest_has_alpha,
                                      const uint8  *src_buf,
                                      int src_width,
                                      int src_height,
                                      int src_rowstride,
                                      int src_channels,
                                      boolean src_has_alpha,
                                      double scale_x,
                                      double scale_y,
                                      int overall_alpha,
                                      int check_x,
                                      int check_y,
                                      int check_size,
                                      uint32 color1,
                                      uint32 color2)
{
  int i, j;
  int x;
  int x_step = (1 << SCALE_SHIFT) / scale_x;
  int y_step = (1 << SCALE_SHIFT) / scale_y;
  int r1, g1, b1, r2, g2, b2;
  int check_shift = get_check_shift (check_size);

  for (i = 0; i < (render_y1 - render_y0); i++) {

     const uint8 *src  = src_buf +
        (((i + render_y0) * y_step + y_step/2) >> SCALE_SHIFT)
        * src_rowstride;
     uint8 *dest = dest_buf + i * dest_rowstride;

     x = render_x0 * x_step + x_step / 2;

     if (((i + check_y) >> check_shift) & 1) {
        r1 = (color2 & 0xff0000) >> 16;
        g1 = (color2 & 0xff00) >> 8;
        b1 = color2 & 0xff;

        r2 = (color1 & 0xff0000) >> 16;
        g2 = (color1 & 0xff00) >> 8;
        b2 = color1 & 0xff;
     } else {
        r1 = (color1 & 0xff0000) >> 16;
        g1 = (color1 & 0xff00) >> 8;
        b1 = color1 & 0xff;

        r2 = (color2 & 0xff0000) >> 16;
        g2 = (color2 & 0xff00) >> 8;
        b2 = color2 & 0xff;
     }

     for (j=0 ; j < (render_x1 - render_x0); j++) {

        const uint8 *p = src + (x >> SCALE_SHIFT) * src_channels;
        int a0;
        int tmp;

        if (src_has_alpha)
           a0 = (p[3] * overall_alpha + 0xff) >> 8;
        else
           a0 = overall_alpha;

        switch (a0) {
        case 0:
           if (((j + check_x) >> check_shift) & 1) {
              dest[0] = r2;
              dest[1] = g2;
              dest[2] = b2;
           } else {
              dest[0] = r1;
              dest[1] = g1;
              dest[2] = b1;
           }
           break;
        case 255:
           dest[0] = p[0];
           dest[1] = p[1];
           dest[2] = p[2];
           break;
        default:
           if (((j + check_x) >> check_shift) & 1) {
              tmp = ((int) p[0] - r2) * a0;
              dest[0] = r2 + ((tmp + (tmp >> 8) + 0x80) >> 8);
              tmp = ((int) p[1] - g2) * a0;
              dest[1] = g2 + ((tmp + (tmp >> 8) + 0x80) >> 8);
              tmp = ((int) p[2] - b2) * a0;
              dest[2] = b2 + ((tmp + (tmp >> 8) + 0x80) >> 8);
           } else {
              tmp = ((int) p[0] - r1) * a0;
              dest[0] = r1 + ((tmp + (tmp >> 8) + 0x80) >> 8);
              tmp = ((int) p[1] - g1) * a0;
              dest[1] = g1 + ((tmp + (tmp >> 8) + 0x80) >> 8);
              tmp = ((int) p[2] - b1) * a0;
              dest[2] = b1 + ((tmp + (tmp >> 8) + 0x80) >> 8);
           }
           break;
        }

        if (dest_channels == 4)
           dest[3] = 0xff;

        dest += dest_channels;
        x += x_step;
     }
  }
}


static void
composite_pixel (uint8 *dest,
                 int dest_x,
                 int dest_bytes_per_pixel,
                 int dest_depth,
                 int dest_has_alpha,
		 int src_has_alpha,
                 int check_size,
                 uint32 color1,
                 uint32 color2,
		 uint r, uint g, uint b, uint a)
{
  if (dest_has_alpha) {
     unsigned int w0 = a - (a >> 8);
     unsigned int w1 = ((0xff0000 - a) >> 8) * dest[3];
     unsigned int w = w0 + w1;

     if (w != 0) {
        dest[0] = (r - (r >> 8) + w1 * dest[0]) / w;
        dest[1] = (g - (g >> 8) + w1 * dest[1]) / w;
        dest[2] = (b - (b >> 8) + w1 * dest[2]) / w;
        dest[3] = w / 0xff00;
     } else {
        dest[0] = 0;
        dest[1] = 0;
        dest[2] = 0;
        dest[3] = 0;
     }
  } else {
     dest[0] = (r + (0xff0000 - a) * dest[0]) / 0xff0000;
     dest[1] = (g + (0xff0000 - a) * dest[1]) / 0xff0000;
     dest[2] = (b + (0xff0000 - a) * dest[2]) / 0xff0000;
  }
}


static uint8 *
composite_line (int *weights,
                int n_x,
                int n_y,
		uint8 *dest,
                int dest_x,
                uint8 *dest_end,
                int dest_bytes_per_pixel,
                int dest_depth,
                int dest_has_alpha,
		uint8 **src,
                int src_bytes_per_pixel,
                int src_depth,
                uint32 *src_colormap,
                boolean src_has_alpha,
		int x_init,
                int x_step,
                int src_width,
		int check_size,
                uint32 color1,
                uint32 color2)
{
   int x = x_init;
   int i, j;

   while (dest < dest_end) {

      int x_scaled = x >> SCALE_SHIFT;
      unsigned int r = 0, g = 0, b = 0, a = 0;
      int *pixel_weights;

      pixel_weights = weights
         + ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK)
         * n_x * n_y;

      for (i=0; i<n_y; i++) {

         uint8 *q = src[i] + x_scaled * src_bytes_per_pixel;
         int *line_weights = pixel_weights + n_x * i;

         for (j=0; j<n_x; j++) {

            unsigned int ta;

            if (src_has_alpha)
               ta = q[3] * line_weights[j];
            else
               ta = 0xff * line_weights[j];

            r += ta * q[0];
            g += ta * q[1];
            b += ta * q[2];
            a += ta;

            q += src_bytes_per_pixel;
         }
      }

      if (dest_has_alpha) {

         unsigned int w0 = a - (a >> 8);
         unsigned int w1 = ((0xff0000 - a) >> 8) * dest[3];
         unsigned int w = w0 + w1;

         if (w != 0) {
            dest[0] = (r - (r >> 8) + w1 * dest[0]) / w;
            dest[1] = (g - (g >> 8) + w1 * dest[1]) / w;
            dest[2] = (b - (b >> 8) + w1 * dest[2]) / w;
            dest[3] = w / 0xff00;
         } else {
            dest[0] = 0;
            dest[1] = 0;
            dest[2] = 0;
            dest[3] = 0;
         }
      } else {
         dest[0] = (r + (0xff0000 - a) * dest[0]) / 0xff0000;
         dest[1] = (g + (0xff0000 - a) * dest[1]) / 0xff0000;
         dest[2] = (b + (0xff0000 - a) * dest[2]) / 0xff0000;
      }

      dest += dest_bytes_per_pixel;
      x += x_step;
   }

   return dest;
}


static uint8 *
composite_line_22_4a4 (int *weights,
                       int n_x,
                       int n_y,
		       uint8 *dest,
                       int dest_x,
                       uint8 *dest_end,
                       int dest_bytes_per_pixel,
                       int dest_depth,
                       int dest_has_alpha,
		       uint8 **src,
                       int src_bytes_per_pixel,
                       int src_depth,
                       uint32 *src_colormap,
                       boolean src_has_alpha,
		       int x_init,
                       int x_step,
                       int src_width,
		       int check_size,
                       uint32 color1,
                       uint32 color2)
{
  int x = x_init;
  uint8 *src0 = src[0];
  uint8 *src1 = src[1];

  return_val_if_fail (src_bytes_per_pixel != 3, dest);
  return_val_if_fail (src_has_alpha, dest);

  while (dest < dest_end) {
     int x_scaled = x >> SCALE_SHIFT;
     unsigned int r, g, b, a, ta;
     int *pixel_weights;
     uint8 *q0, *q1;
     int w1, w2, w3, w4;

     q0 = src0 + x_scaled * 4;
     q1 = src1 + x_scaled * 4;

     pixel_weights = (int *)((char *)weights + ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS - 4)) & (SUBSAMPLE_MASK << 4)));

     w1 = pixel_weights[0];
     w2 = pixel_weights[1];
     w3 = pixel_weights[2];
     w4 = pixel_weights[3];

     a = w1 * q0[3];
     r = a * q0[0];
     g = a * q0[1];
     b = a * q0[2];

     ta = w2 * q0[7];
     r += ta * q0[4];
     g += ta * q0[5];
     b += ta * q0[6];
     a += ta;

     ta = w3 * q1[3];
     r += ta * q1[0];
     g += ta * q1[1];
     b += ta * q1[2];
     a += ta;

     ta = w4 * q1[7];
     r += ta * q1[4];
     g += ta * q1[5];
     b += ta * q1[6];
     a += ta;

     dest[0] = ((0xff0000 - a) * dest[0] + r) >> 24;
     dest[1] = ((0xff0000 - a) * dest[1] + g) >> 24;
     dest[2] = ((0xff0000 - a) * dest[2] + b) >> 24;
     dest[3] = a >> 16;

     dest += 4;
     x += x_step;
  }

  return dest;
}


#ifdef USE_MMX
static uint8 *
composite_line_22_4a4_mmx_stub (int *weights,
                                int n_x,
                                int n_y,
				uint8 *dest,
                                int dest_x,
                                uint8 *dest_end,
                                int dest_bytes_per_pixel,
                                int dest_depth,
                                int dest_has_alpha,
				uint8 **src,
                                int src_bytes_per_pixel,
                                int src_depth,
                                uint32 *src_colormap,
                                boolean src_has_alpha,
				int x_init,
                                int x_step,
                                int src_width,
				int check_size,
                                uint32 color1,
                                uint32 color2)
{
   uint32 mmx_weights[16][8];
   int j;

   for (j=0; j<16; j++) {
      mmx_weights[j][0] = 0x00010001 * (weights[4*j] >> 8);
      mmx_weights[j][1] = 0x00010001 * (weights[4*j] >> 8);
      mmx_weights[j][2] = 0x00010001 * (weights[4*j + 1] >> 8);
      mmx_weights[j][3] = 0x00010001 * (weights[4*j + 1] >> 8);
      mmx_weights[j][4] = 0x00010001 * (weights[4*j + 2] >> 8);
      mmx_weights[j][5] = 0x00010001 * (weights[4*j + 2] >> 8);
      mmx_weights[j][6] = 0x00010001 * (weights[4*j + 3] >> 8);
      mmx_weights[j][7] = 0x00010001 * (weights[4*j + 3] >> 8);
   }

   return pixops_composite_line_22_4a4_mmx (mmx_weights, dest, src[0], src[1],
                                            x_step, dest_end, x_init);
}
#endif /* USE_MMX */


static void
composite_pixel_color (uint8 *dest,
                       int dest_x,
                       int dest_bytes_per_pixel,
                       int dest_depth,
                       int dest_has_alpha,
		       int src_has_alpha,
                       int check_size,
                       uint32 color1,
                       uint32 color2,
		       uint r, uint g, uint b, uint a)
{
   int dest_r, dest_g, dest_b;
   int check_shift = get_check_shift (check_size);

   if ((dest_x >> check_shift) & 1) {
      dest_r = (color2 & 0xff0000) >> 16;
      dest_g = (color2 & 0xff00) >> 8;
      dest_b = color2 & 0xff;
   } else {
      dest_r = (color1 & 0xff0000) >> 16;
      dest_g = (color1 & 0xff00) >> 8;
      dest_b = color1 & 0xff;
   }

   dest[0] = ((0xff0000 - a) * dest_r + r) >> 24;
   dest[1] = ((0xff0000 - a) * dest_g + g) >> 24;
   dest[2] = ((0xff0000 - a) * dest_b + b) >> 24;

   if (dest_has_alpha)
      dest[3] = 0xff;
   else if (dest_bytes_per_pixel == 4)
      dest[3] = a >> 16;
}


static uint8 *
composite_line_color (int *weights,
                      int n_x,
                      int n_y,
                      uint8 *dest,
                      int dest_x,
                      uint8 *dest_end,
                      int dest_bytes_per_pixel,
                      int dest_depth,
                      int dest_has_alpha,
                      uint8 **src,
                      int src_bytes_per_pixel,
                      int src_depth,
                      uint32 *src_colormap,
                      boolean src_has_alpha,
                      int x_init,
                      int x_step,
                      int src_width,
                      int check_size,
                      uint32 color1,
                      uint32 color2)
{
   int x = x_init;
   int i, j;
   int check_shift = get_check_shift (check_size);
   int dest_r1, dest_g1, dest_b1;
   int dest_r2, dest_g2, dest_b2;

   return_val_if_fail (check_size != 0, dest);

   dest_r1 = (color1 & 0xff0000) >> 16;
   dest_g1 = (color1 & 0xff00) >> 8;
   dest_b1 = color1 & 0xff;

   dest_r2 = (color2 & 0xff0000) >> 16;
   dest_g2 = (color2 & 0xff00) >> 8;
   dest_b2 = color2 & 0xff;

   while (dest < dest_end) {
      int x_scaled = x >> SCALE_SHIFT;
      unsigned int r = 0, g = 0, b = 0, a = 0;
      int *pixel_weights;

      pixel_weights = weights + ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK) * n_x * n_y;

      for (i=0; i<n_y; i++) {
         uint8 *q = src[i] + x_scaled * src_bytes_per_pixel;
         int *line_weights = pixel_weights + n_x * i;

         for (j=0; j<n_x; j++) {
            unsigned int ta;

            if (src_has_alpha)
               ta = q[3] * line_weights[j];
            else
               ta = 0xff * line_weights[j];

            r += ta * q[0];
            g += ta * q[1];
            b += ta * q[2];
            a += ta;

            q += src_bytes_per_pixel;
         }
      }

      if ((dest_x >> check_shift) & 1) {
         dest[0] = ((0xff0000 - a) * dest_r2 + r) >> 24;
         dest[1] = ((0xff0000 - a) * dest_g2 + g) >> 24;
         dest[2] = ((0xff0000 - a) * dest_b2 + b) >> 24;
      } else {
         dest[0] = ((0xff0000 - a) * dest_r1 + r) >> 24;
         dest[1] = ((0xff0000 - a) * dest_g1 + g) >> 24;
         dest[2] = ((0xff0000 - a) * dest_b1 + b) >> 24;
      }

      if (dest_has_alpha)
         dest[3] = 0xff;
      else if (dest_bytes_per_pixel == 4)
         dest[3] = a >> 16;

      dest += dest_bytes_per_pixel;
      x += x_step;
      dest_x++;
   }

   return dest;
}


#ifdef USE_MMX
static uint8 *
composite_line_color_22_4a4_mmx_stub (int *weights,
                                      int n_x,
                                      int n_y,
                                      uint8 *dest,
                                      int dest_x,
                                      uint8 *dest_end,
                                      int dest_bytes_per_pixel,
                                      int dest_depth,
                                      int dest_has_alpha,
                                      uint8 **src,
                                      int src_bytes_per_pixel,
                                      int src_depth,
                                      uint32 *src_colormap,
                                      boolean src_has_alpha,
                                      int x_init,
                                      int x_step,
                                      int src_width,
                                      int check_size,
                                      uint32 color1,
                                      uint32 color2)
{
   uint32 mmx_weights[16][8];
   int check_shift = get_check_shift (check_size);
   int colors[4];
   int j;

   for (j=0; j<16; j++) {
      mmx_weights[j][0] = 0x00010001 * (weights[4*j] >> 8);
      mmx_weights[j][1] = 0x00010001 * (weights[4*j] >> 8);
      mmx_weights[j][2] = 0x00010001 * (weights[4*j + 1] >> 8);
      mmx_weights[j][3] = 0x00010001 * (weights[4*j + 1] >> 8);
      mmx_weights[j][4] = 0x00010001 * (weights[4*j + 2] >> 8);
      mmx_weights[j][5] = 0x00010001 * (weights[4*j + 2] >> 8);
      mmx_weights[j][6] = 0x00010001 * (weights[4*j + 3] >> 8);
      mmx_weights[j][7] = 0x00010001 * (weights[4*j + 3] >> 8);
   }

   colors[0] = (color1 & 0xff00) << 8 | (color1 & 0xff);
   colors[1] = (color1 & 0xff0000) >> 16;
   colors[2] = (color2 & 0xff00) << 8 | (color2 & 0xff);
   colors[3] = (color2 & 0xff0000) >> 16;

   return pixops_composite_line_color_22_4a4_mmx (mmx_weights, dest, src[0], src[1],
                                                  x_step, dest_end, x_init,
                                                  dest_x, check_shift, colors);
}
#endif /* USE_MMX */


static void
scale_pixel (uint8 *dest,
             int dest_x,
             int dest_bytes_per_pixel,
             int dest_depth,
             int dest_has_alpha,
             int src_has_alpha,
             int check_size,
             uint32 color1,
             uint32 color2,
             uint r, uint g, uint b, uint a)
{
   uint32 p;

   if (src_has_alpha) {
      if (a) {
         ((uint8 *) &p)[0] = r / a;
         ((uint8 *) &p)[1] = g / a;
         ((uint8 *) &p)[2] = b / a;
         ((uint8 *) &p)[3] = a >> 16;
      } else {
         ((uint8 *) &p)[0] = 0;
         ((uint8 *) &p)[1] = 0;
         ((uint8 *) &p)[2] = 0;
         ((uint8 *) &p)[3] = 0;
      }
   } else {
      ((uint8 *) &p)[0] = (r + 0xffffff) >> 24;
      ((uint8 *) &p)[1] = (g + 0xffffff) >> 24;
      ((uint8 *) &p)[2] = (b + 0xffffff) >> 24;

      if (dest_has_alpha)
         ((uint8 *) &p)[3] = 0xff;
   }

   store_pixel (dest, dest_bytes_per_pixel, dest_depth, p);
}


static uint8 *
scale_line (int *weights,
            int n_x,
            int n_y,
            uint8 *dest,
            int dest_x,
            uint8 *dest_end,
            int dest_bytes_per_pixel,
            int dest_depth,
            int dest_has_alpha,
            uint8 **src,
            int src_bytes_per_pixel,
            int src_depth,
            uint32 *src_colormap,
            boolean src_has_alpha,
            int x_init,
            int x_step,
            int src_width,
            int check_size,
            uint32 color1,
            uint32 color2)
{
   int x = x_init;
   int i, j;
   int src_bppdepth = BPPDEPTH(src_bytes_per_pixel << 3, src_depth);

   while (dest < dest_end) {
      int x_scaled = x >> SCALE_SHIFT;
      int *pixel_weights;
      uint32 p;

      pixel_weights = weights
         + ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK)
         * n_x * n_y;

      if (src_has_alpha) {
         unsigned int r = 0, g = 0, b = 0, a = 0;
         for (i=0; i<n_y; i++) {
            uint8 *q = src[i] + x_scaled * src_bytes_per_pixel;
            int *line_weights  = pixel_weights + n_x * i;

            for (j=0; j<n_x; j++) {
               unsigned int ta;

               ta = q[3] * line_weights[j];
               r += ta * q[0];
               g += ta * q[1];
               b += ta * q[2];
               a += ta;

               q += src_bytes_per_pixel;
            }
         }

         if (a) {
            ((uint8 *) &p)[0] = r / a;
            ((uint8 *) &p)[1] = g / a;
            ((uint8 *) &p)[2] = b / a;
            ((uint8 *) &p)[3] = a >> 16;
         } else {
            ((uint8 *) &p)[0] = 0;
            ((uint8 *) &p)[1] = 0;
            ((uint8 *) &p)[2] = 0;
            ((uint8 *) &p)[3] = 0;
         }
      } else {
         unsigned int r = 0, g = 0, b = 0;
         for (i=0; i<n_y; i++) {
            uint8 *q = src[i] + x_scaled * src_bytes_per_pixel;
            int *line_weights  = pixel_weights + n_x * i;

            for (j=0; j<n_x; j++) {
               unsigned int ta = line_weights[j];
               uint32 tp;

               tp = load_pixel (q, src_bytes_per_pixel, src_depth, src_colormap);
               r += ta * ((uint8 *) &tp)[0];
               g += ta * ((uint8 *) &tp)[1];
               b += ta * ((uint8 *) &tp)[2];

               q += src_bytes_per_pixel;
            }
         }

         ((uint8 *) &p)[0] = (r + 0xffff) >> 16;
         ((uint8 *) &p)[1] = (g + 0xffff) >> 16;
         ((uint8 *) &p)[2] = (b + 0xffff) >> 16;

         if (dest_has_alpha)
            ((uint8 *) &p)[3] = 0xff;
      }

      store_pixel (dest, dest_bytes_per_pixel, dest_depth, p);
      dest += dest_bytes_per_pixel;
      x += x_step;
   }

   return dest;
}


#ifdef USE_MMX
static uint8 *
scale_line_22_33_mmx_stub (int *weights,
                           int n_x,
                           int n_y,
                           uint8 *dest,
                           int dest_x,
                           uint8 *dest_end,
                           int dest_bytes_per_pixel,
                           int dest_depth,
                           int dest_has_alpha,
                           uint8 **src,
                           int src_bytes_per_pixel,
                           int src_depth,
                           uint32 *src_colormap,
                           boolean src_has_alpha,
                           int x_init,
                           int x_step,
                           int src_width,
                           int check_size,
                           uint32 color1,
                           uint32 color2)
{
   uint32 mmx_weights[16][8];
   int j;

   for (j=0; j<16; j++) {
      mmx_weights[j][0] = 0x00010001 * (weights[4*j] >> 8);
      mmx_weights[j][1] = 0x00010001 * (weights[4*j] >> 8);
      mmx_weights[j][2] = 0x00010001 * (weights[4*j + 1] >> 8);
      mmx_weights[j][3] = 0x00010001 * (weights[4*j + 1] >> 8);
      mmx_weights[j][4] = 0x00010001 * (weights[4*j + 2] >> 8);
      mmx_weights[j][5] = 0x00010001 * (weights[4*j + 2] >> 8);
      mmx_weights[j][6] = 0x00010001 * (weights[4*j + 3] >> 8);
      mmx_weights[j][7] = 0x00010001 * (weights[4*j + 3] >> 8);
   }

   return pixops_scale_line_22_33_mmx (mmx_weights, dest, src[0], src[1], x_step, dest_end, x_init);
}
#endif /* USE_MMX */


static uint8 *
scale_line_22_33 (int *weights,
                  int n_x,
                  int n_y,
                  uint8 *dest,
                  int dest_x,
                  uint8 *dest_end,
                  int dest_bytes_per_pixel,
                  int dest_depth,
                  int dest_has_alpha,
                  uint8 **src,
                  int src_bytes_per_pixel,
                  int src_depth,
                  uint32 *src_colormap,
                  boolean src_has_alpha,
                  int x_init,
                  int x_step,
                  int src_width,
                  int check_size,
                  uint32 color1,
                  uint32 color2)
{
   int x = x_init;
   uint8 *src0 = src[0];
   uint8 *src1 = src[1];

   while (dest < dest_end) {
      unsigned int r, g, b;
      int x_scaled = x >> SCALE_SHIFT;
      int *pixel_weights;
      uint8 *q0, *q1;
      int w1, w2, w3, w4;

      q0 = src0 + x_scaled * 3;
      q1 = src1 + x_scaled * 3;

      pixel_weights = weights + ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK) * 4;

      w1 = pixel_weights[0];
      w2 = pixel_weights[1];
      w3 = pixel_weights[2];
      w4 = pixel_weights[3];

      r = w1 * q0[0];
      g = w1 * q0[1];
      b = w1 * q0[2];

      r += w2 * q0[3];
      g += w2 * q0[4];
      b += w2 * q0[5];

      r += w3 * q1[0];
      g += w3 * q1[1];
      b += w3 * q1[2];

      r += w4 * q1[3];
      g += w4 * q1[4];
      b += w4 * q1[5];

      dest[0] = (r + 0x8000) >> 16;
      dest[1] = (g + 0x8000) >> 16;
      dest[2] = (b + 0x8000) >> 16;

      dest += 3;
      x += x_step;
   }

   return dest;
}


static void
process_pixel (int *weights,
               int n_x,
               int n_y,
               uint8 *dest,
               int dest_x,
               int dest_bytes_per_pixel,
               int dest_depth,
               int dest_has_alpha,
               uint8 **src,
               int src_bytes_per_pixel,
               int src_depth,
               uint32 *src_colormap,
               boolean src_has_alpha,
               int x_start,
               int src_width,
               int check_size,
               uint32 color1,
               uint32 color2,
               PixopsPixelFunc pixel_func)
{
   unsigned int r = 0, g = 0, b = 0, a = 0;
   int i, j;

   for (i=0; i<n_y; i++)
   {
      int *line_weights  = weights + n_x * i;

      for (j=0; j<n_x; j++)
      {
         unsigned int ta;
         uint8 *q;
         uint32 p;

         if (x_start + j < 0)
            q = src[i];
         else if (x_start + j < src_width)
            q = src[i] + (x_start + j) * src_bytes_per_pixel;
         else
            q = src[i] + (src_width - 1) * src_bytes_per_pixel;

         if (src_has_alpha)
            ta = q[3] * line_weights[j];
         else
            ta = 0xff * line_weights[j];

         p = load_pixel (q, src_bytes_per_pixel, src_depth, src_colormap);
         r += ta * (((uint8 *) &p)[0]);
         g += ta * (((uint8 *) &p)[1]);
         b += ta * (((uint8 *) &p)[2]);
         a += ta;
      }
   }

   (*pixel_func) (dest, dest_x, dest_bytes_per_pixel, dest_depth, dest_has_alpha,
                  src_has_alpha, check_size, color1, color2, r, g, b, a);
}


static void
pixops_process (uint8 *dest_buf,
                int render_x0,
                int render_y0,
                int render_x1,
                int render_y1,
                int dest_rowstride,
                int dest_bytes_per_pixel,
                int dest_depth,
                boolean dest_has_alpha,
                const uint8   *src_buf,
                int src_width,
                int src_height,
                int src_rowstride,
                int src_bytes_per_pixel,
                int src_depth,
                uint32 *src_colormap,
                boolean src_has_alpha,
                double scale_x,
                double scale_y,
                int check_x,
                int check_y,
                int check_size,
                uint32 color1,
                uint32 color2,
                PixopsFilter *filter,
                PixopsLineFunc line_func,
                PixopsPixelFunc pixel_func)
{
   int i, j;
   int x, y;			/* X and Y position in source (fixed_point) */
   uint8 **line_bufs = mnew (uint8 *, filter->n_y);

   int x_step = (1 << SCALE_SHIFT) / scale_x; /* X step in source (fixed point) */
   int y_step = (1 << SCALE_SHIFT) / scale_y; /* Y step in source (fixed point) */

   int check_shift = check_size ? get_check_shift (check_size) : 0;

   int scaled_x_offset = floor (filter->x_offset * (1 << SCALE_SHIFT));

   /* Compute the index where we run off the end of the source buffer. The
    * furthest source pixel we access at index i is:
    *
    *  ((render_x0 + i) * x_step + scaled_x_offset) >> SCALE_SHIFT + filter->n_x - 1
    *
    * So, run_end_index is the smallest i for which this pixel is src_width,
    * i.e, for which:
    *
    *  (i + render_x0) * x_step >= ((src_width - filter->n_x + 1) << SCALE_SHIFT) - scaled_x_offset
    *
    */
#define MYDIV(a,b) ((a) > 0 ? (a) / (b) : ((a) - (b) + 1) / (b))    /* Division so that -1/5 = -1 */

   int run_end_x = (((src_width - filter->n_x + 1) << SCALE_SHIFT) - scaled_x_offset);
   int run_end_index = MYDIV (run_end_x + x_step - 1, x_step) - render_x0;
   run_end_index = MIN (run_end_index, render_x1 - render_x0);

   y = render_y0 * y_step + floor (filter->y_offset * (1 << SCALE_SHIFT));
   for (i = 0; i < (render_y1 - render_y0); i++)
   {
      int dest_x;
      int y_start = y >> SCALE_SHIFT;
      int x_start;
      int *run_weights = filter->weights
         + ((y >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK)
         * filter->n_x * filter->n_y
         * SUBSAMPLE;
      uint8 *new_outbuf;
      uint32 tcolor1, tcolor2;

      uint8 *outbuf = dest_buf + dest_rowstride * i;
      uint8 *outbuf_end = outbuf + dest_bytes_per_pixel * (render_x1 - render_x0);

      if (((i + check_y) >> check_shift) & 1)
      {
         tcolor1 = color2;
         tcolor2 = color1;
      }
      else
      {
         tcolor1 = color1;
         tcolor2 = color2;
      }

      for (j=0; j<filter->n_y; j++)
      {
         if (y_start <  0)
            line_bufs[j] = (uint8 *)src_buf;
         else if (y_start < src_height)
            line_bufs[j] = (uint8 *)src_buf + src_rowstride * y_start;
         else
            line_bufs[j] = (uint8 *)src_buf + src_rowstride * (src_height - 1);

         y_start++;
      }

      dest_x = check_x;
      x = render_x0 * x_step + scaled_x_offset;
      x_start = x >> SCALE_SHIFT;

      while (x_start < 0 && outbuf < outbuf_end)
      {
         process_pixel (run_weights
                         + ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK)
                         * (filter->n_x * filter->n_y),
                        filter->n_x, filter->n_y,
                        outbuf, dest_x, dest_bytes_per_pixel, dest_depth, dest_has_alpha,
                        line_bufs, src_bytes_per_pixel, src_depth, src_colormap,
                        src_has_alpha, x >> SCALE_SHIFT, src_width,
                        check_size, tcolor1, tcolor2, pixel_func);

         x += x_step;
         x_start = x >> SCALE_SHIFT;
         dest_x++;
         outbuf += dest_bytes_per_pixel;
      }

      new_outbuf = (*line_func) (run_weights, filter->n_x, filter->n_y,
                                 outbuf, dest_x,
                                 dest_buf + dest_rowstride * i
                                  + run_end_index * dest_bytes_per_pixel,
                                 dest_bytes_per_pixel, dest_depth, dest_has_alpha,
                                 line_bufs, src_bytes_per_pixel, src_depth,
                                 src_colormap, src_has_alpha,
                                 x, x_step, src_width, check_size, tcolor1, tcolor2);

      dest_x += (new_outbuf - outbuf) / dest_bytes_per_pixel;

      x = (dest_x - check_x + render_x0) * x_step + scaled_x_offset;
      outbuf = new_outbuf;

      while (outbuf < outbuf_end)
      {
         process_pixel (run_weights +
                         ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK)
                         * (filter->n_x * filter->n_y),
                        filter->n_x, filter->n_y,
                        outbuf, dest_x, dest_bytes_per_pixel,
                        dest_depth, dest_has_alpha,
                        line_bufs, src_bytes_per_pixel, src_depth, src_colormap,
                        src_has_alpha, x >> SCALE_SHIFT, src_width,
                        check_size, tcolor1, tcolor2, pixel_func);

         x += x_step;
         dest_x++;
         outbuf += dest_bytes_per_pixel;
      }

      y += y_step;
   }

   free (line_bufs);
}


static void
correct_total (int    *weights,
               int    n_x,
               int    n_y,
               int    total,
               double overall_alpha)
{
   int correction = (int)(0.5 + 65536 * overall_alpha) - total;
   int i;
   for (i = n_x * n_y - 1; i >= 0; i--) {
      if (*(weights + i) + correction >= 0) {
         *(weights + i) += correction;
         break;
      }
   }
}


static void
tile_make_weights (PixopsFilter *filter, double x_scale, double y_scale, double overall_alpha)
{
   int i_offset, j_offset;

   int n_x = ceil(1/x_scale + 1);
   int n_y = ceil(1/y_scale + 1);

   filter->x_offset = 0;
   filter->y_offset = 0;
   filter->n_x = n_x;
   filter->n_y = n_y;
   filter->weights = mnew (int, SUBSAMPLE * SUBSAMPLE * n_x * n_y);

   for (i_offset=0; i_offset<SUBSAMPLE; i_offset++) {
      for (j_offset=0; j_offset<SUBSAMPLE; j_offset++)
      {
         int *pixel_weights = filter->weights + ((i_offset*SUBSAMPLE) + j_offset) * n_x * n_y;
         double x = (double)j_offset / SUBSAMPLE;
         double y = (double)i_offset / SUBSAMPLE;
         int i,j;
         int total = 0;

         for (i = 0; i < n_y; i++)
         {
            double tw, th;

            if (i < y)
            {

               if (i + 1 > y)
                  th = MIN(i+1, y + 1/y_scale) - y;
               else
                  th = 0;
            }
            else
            {
               if (y + 1/y_scale > i)
                  th = MIN(i+1, y + 1/y_scale) - i;
               else
                  th = 0;
            }

            for (j = 0; j < n_x; j++)
            {
               int weight;

               if (j < x)
               {
                  if (j + 1 > x)
                     tw = MIN(j+1, x + 1/x_scale) - x;
                  else
                     tw = 0;
               }
               else
               {
                  if (x + 1/x_scale > j)
                     tw = MIN(j+1, x + 1/x_scale) - j;
                  else
                     tw = 0;
               }

               weight = 65536 * tw * x_scale * th * y_scale * overall_alpha + 0.5;
               total += weight;
               *(pixel_weights + n_x * i + j) = weight;
            }
         }

         correct_total (pixel_weights, n_x, n_y, total, overall_alpha);
      }
   }
}


static void
bilinear_make_fast_weights (PixopsFilter *filter,
                            double x_scale,
                            double y_scale,
                            double overall_alpha)
{
   int i_offset, j_offset;
   double *x_weights, *y_weights;
   int n_x, n_y;

   if (x_scale > 1.0)		/* Bilinear */
   {
      n_x = 2;
      filter->x_offset = 0.5 * (1/x_scale - 1);
   }
   else				/* Tile */
   {
      n_x = ceil(1.0 + 1.0/x_scale);
      filter->x_offset = 0.0;
   }

   if (y_scale > 1.0)		/* Bilinear */
   {
      n_y = 2;
      filter->y_offset = 0.5 * (1/y_scale - 1);
   }
   else				/* Tile */
   {
      n_y = ceil(1.0 + 1.0/y_scale);
      filter->y_offset = 0.0;
   }

   filter->n_y = n_y;
   filter->n_x = n_x;
   filter->weights = mnew (int, SUBSAMPLE * SUBSAMPLE * n_x * n_y);

   x_weights = mnew (double, n_x);
   y_weights = mnew (double, n_y);

   for (i_offset=0; i_offset<SUBSAMPLE; i_offset++) {
      for (j_offset=0; j_offset<SUBSAMPLE; j_offset++)
      {
         int *pixel_weights = filter->weights + ((i_offset*SUBSAMPLE) + j_offset) * n_x * n_y;
         double x = (double)j_offset / SUBSAMPLE;
         double y = (double)i_offset / SUBSAMPLE;
         int i,j;
         int total = 0;

         if (x_scale > 1.0)	/* Bilinear */
         {
            for (i = 0; i < n_x; i++)
            {
               x_weights[i] = ((i == 0) ? (1 - x) : x) / x_scale;
            }
         }
         else			/* Tile */
         {
            /*           x
             * ---------|--.-|----|--.-|-------  SRC
             * ------------|---------|---------  DEST
             */
            for (i = 0; i < n_x; i++)
            {
               if (i < x)
               {
                  if (i + 1 > x)
                     x_weights[i] = MIN(i+1, x + 1/x_scale) - x;
                  else
                     x_weights[i] = 0;
               }
               else
               {
                  if (x + 1/x_scale > i)
                     x_weights[i] = MIN(i+1, x + 1/x_scale) - i;
                  else
                     x_weights[i] = 0;
               }
            }
         }

         if (y_scale > 1.0)	/* Bilinear */
         {
            for (i = 0; i < n_y; i++)
            {
               y_weights[i] = ((i == 0) ? (1 - y) : y) / y_scale;
            }
         }
         else			/* Tile */
         {
            /*           y
             * ---------|--.-|----|--.-|-------  SRC
             * ------------|---------|---------  DEST
             */
            for (i = 0; i < n_y; i++)
            {
               if (i < y)
               {
                  if (i + 1 > y)
                     y_weights[i] = MIN(i+1, y + 1/y_scale) - y;
                  else
                     y_weights[i] = 0;
               }
               else
               {
                  if (y + 1/y_scale > i)
                     y_weights[i] = MIN(i+1, y + 1/y_scale) - i;
                  else
                     y_weights[i] = 0;
               }
            }
         }

         for (i = 0; i < n_y; i++) {
            for (j = 0; j < n_x; j++) {

               int weight = 65536 *
                  x_weights[j] * x_scale *
                  y_weights[i] * y_scale *
                  overall_alpha + 0.5;

               *(pixel_weights + n_x * i + j) = weight;
               total += weight;
            }
         }

         correct_total (pixel_weights, n_x, n_y, total, overall_alpha);
      }
   }

   free (x_weights);
   free (y_weights);
}


static double
bilinear_quadrant (double bx0,
                   double bx1,
                   double by0,
                   double by1)
{
   double ax0, ax1, ay0, ay1;
   double x0, x1, y0, y1;

   ax0 = 0.;
   ax1 = 1.;
   ay0 = 0.;
   ay1 = 1.;

   if (ax0 < bx0) {
      if (ax1 > bx0) {
         x0 = bx0;
         x1 = MIN (ax1, bx1);
      } else {
         return 0;
      }
   } else {
      if (bx1 > ax0) {
         x0 = ax0;
         x1 = MIN (ax1, bx1);
      } else {
         return 0;
      }
   }

   if (ay0 < by0) {
      if (ay1 > by0) {
         y0 = by0;
         y1 = MIN (ay1, by1);
      } else {
         return 0;
      }
   } else {
      if (by1 > ay0) {
         y0 = ay0;
         y1 = MIN (ay1, by1);
      } else {
         return 0;
      }
   }

   return 0.25 * (x1*x1 - x0*x0) * (y1*y1 - y0*y0);
}


static void
bilinear_make_weights (PixopsFilter *filter,
                       double x_scale,
                       double y_scale,
                       double overall_alpha)
{
   int i_offset, j_offset;

   int n_x = ceil(1/x_scale + 2.0);
   int n_y = ceil(1/y_scale + 2.0);

   filter->x_offset = -1.0;
   filter->y_offset = -1.0;
   filter->n_x = n_x;
   filter->n_y = n_y;

   filter->weights = mnew (int, SUBSAMPLE * SUBSAMPLE * n_x * n_y);

   for (i_offset=0; i_offset<SUBSAMPLE; i_offset++) {
      for (j_offset=0; j_offset<SUBSAMPLE; j_offset++) {
         int *pixel_weights = filter->weights + ((i_offset*SUBSAMPLE) + j_offset) * n_x * n_y;
         double x = (double)j_offset / SUBSAMPLE;
         double y = (double)i_offset / SUBSAMPLE;
         int i,j;
         int total = 0;

         for (i = 0; i < n_y; i++) {
            for (j = 0; j < n_x; j++) {
               double w;
               int weight;

               w = bilinear_quadrant  (0.5 + j - (x + 1 / x_scale), 0.5 + j - x, 0.5 + i - (y + 1 / y_scale), 0.5 + i - y);
               w += bilinear_quadrant (1.5 + x - j, 1.5 + (x + 1 / x_scale) - j, 0.5 + i - (y + 1 / y_scale), 0.5 + i - y);
               w += bilinear_quadrant (0.5 + j - (x + 1 / x_scale), 0.5 + j - x, 1.5 + y - i, 1.5 + (y + 1 / y_scale) - i);
               w += bilinear_quadrant (1.5 + x - j, 1.5 + (x + 1 / x_scale) - j, 1.5 + y - i, 1.5 + (y + 1 / y_scale) - i);
               weight = 65536 * w * x_scale * y_scale * overall_alpha + 0.5;
               *(pixel_weights + n_x * i + j) = weight;
               total += weight;
            }
         }

         correct_total (pixel_weights, n_x, n_y, total, overall_alpha);
      }
   }
}


PixopsStatus
pixops_scale_composite_color (uint8 *dest_buf,
                              int render_x0,
                              int render_y0,
                              int render_x1,
                              int render_y1,
                              int dest_rowstride,
                              int dest_channels,
                              boolean dest_has_alpha,
                              const uint8   *src_buf,
                              int src_width,
                              int src_height,
                              int src_rowstride,
                              int src_channels,
                              boolean src_has_alpha,
                              double scale_x,
                              double scale_y,
                              PixopsInterpType   interp_type,
                              int overall_alpha,
                              int check_x,
                              int check_y,
                              int check_size,
                              uint32 color1,
                              uint32 color2)
{
   PixopsFilter filter;
   PixopsLineFunc line_func;

#ifdef USE_MMX
   boolean found_mmx = pixops_have_mmx();
#endif

   if (pixops_support_scale_composite_formats (dest_channels, dest_has_alpha,
                                               src_channels, src_has_alpha)
       != PIXOPS_STATUS_SUCCESS)
   {
      return PIXOPS_STATUS_UNSUPPORTED_FORMAT;
   }

   if (scale_x == 0 || scale_y == 0)
      return PIXOPS_STATUS_FAILURE;

   if (!src_has_alpha && overall_alpha == 255) {
      return pixops_scale (dest_buf, render_x0, render_y0, render_x1,
                           render_y1, dest_rowstride, dest_channels, 24,
                           dest_has_alpha, src_buf, src_width, src_height,
                           src_rowstride, src_channels, 24, NULL,
                           src_has_alpha, scale_x, scale_y, interp_type);
   }

   switch (interp_type) {

   case PIXOPS_INTERP_NEAREST:
      pixops_scale_composite_color_nearest (dest_buf, render_x0, render_y0,
                                            render_x1, render_y1,
                                            dest_rowstride, dest_channels,
                                            dest_has_alpha, src_buf,
                                            src_width, src_height,
                                            src_rowstride, src_channels,
                                            src_has_alpha, scale_x, scale_y,
                                            overall_alpha, check_x, check_y,
                                            check_size, color1, color2);
      return PIXOPS_STATUS_SUCCESS;

   case PIXOPS_INTERP_TILES:
      tile_make_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
      break;

   case PIXOPS_INTERP_BILINEAR:
      bilinear_make_fast_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
      break;

   case PIXOPS_INTERP_HYPER:
      bilinear_make_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
      break;
   }

#ifdef USE_MMX
   if (filter.n_x == 2 && filter.n_y == 2 &&
       dest_channels == 4 && src_channels == 4 &&
       src_has_alpha && !dest_has_alpha && found_mmx) {

      line_func = composite_line_color_22_4a4_mmx_stub;
   } else {
#endif
      line_func = composite_line_color;
   }

   pixops_process (dest_buf, render_x0, render_y0, render_x1, render_y1,
                   dest_rowstride, dest_channels, 24, dest_has_alpha, src_buf,
                   src_width, src_height, src_rowstride, src_channels, 24,
                   NULL, src_has_alpha, scale_x, scale_y, check_x, check_y,
                   check_size, color1, color2, &filter, line_func,
                   composite_pixel_color);

   free (filter.weights);
   return PIXOPS_STATUS_SUCCESS;
}


/**
 * pixops_scale_composite:
 * @dest_buf: pointer to location to store result
 * @render_x0: x0 of region of scaled source to store into @dest_buf
 * @render_y0: y0 of region of scaled source to store into @dest_buf
 * @render_x1: x1 of region of scaled source to store into @dest_buf
 * @render_y1: y1 of region of scaled source to store into @dest_buf
 * @dest_rowstride: rowstride of @dest_buf
 * @dest_channels: number of channels in @dest_buf
 * @dest_has_alpha: whether @dest_buf has alpha
 * @src_buf: pointer to source pixels
 * @src_width: width of source (used for clipping)
 * @src_height: height of source (used for clipping)
 * @src_rowstride: rowstride of source
 * @src_channels: number of channels in @src_buf
 * @src_has_alpha: whether @src_buf has alpha
 * @scale_x: amount to scale source by in X direction
 * @scale_y: amount to scale source by in Y direction
 * @interp_type: type of enumeration
 * @overall_alpha: overall alpha factor to multiply source by
 * 
 * Scale source buffer by scale_x / scale_y, then composite a given rectangle
 * of the result into the destination buffer.
 **/
PixopsStatus
pixops_scale_composite (uint8 *dest_buf,
                        int render_x0,
                        int render_y0,
                        int render_x1,
                        int render_y1,
                        int dest_rowstride,
                        int dest_channels,
                        boolean dest_has_alpha,
                        const uint8  *src_buf,
                        int src_width,
                        int src_height,
                        int src_rowstride,
                        int src_channels,
                        boolean src_has_alpha,
                        double scale_x,
                        double scale_y,
                        PixopsInterpType interp_type,
                        int overall_alpha)
{
   PixopsFilter filter;
   PixopsLineFunc line_func;

#ifdef USE_MMX
   boolean found_mmx = pixops_have_mmx();
#endif

   if (pixops_support_scale_composite_formats (dest_channels, dest_has_alpha,
                                               src_channels, src_has_alpha)
       != PIXOPS_STATUS_SUCCESS)
   {
      return PIXOPS_STATUS_UNSUPPORTED_FORMAT;
   }

   if (scale_x == 0 || scale_y == 0) {
      return PIXOPS_STATUS_FAILURE;
   }

   if (!src_has_alpha && overall_alpha == 255)
   {
      return pixops_scale (dest_buf, render_x0, render_y0, render_x1, render_y1,
                           dest_rowstride, dest_channels, 24, dest_has_alpha,
                           src_buf, src_width, src_height, src_rowstride, src_channels,
                           24, NULL, src_has_alpha, scale_x, scale_y, interp_type);
   }

   switch (interp_type) {
   case PIXOPS_INTERP_NEAREST:
      pixops_scale_composite_nearest (dest_buf, render_x0, render_y0,
                                      render_x1, render_y1, dest_rowstride,
                                      dest_channels, dest_has_alpha, src_buf,
                                      src_width, src_height, src_rowstride,
                                      src_channels, src_has_alpha, scale_x,
                                      scale_y, overall_alpha);
      return PIXOPS_STATUS_SUCCESS;

   case PIXOPS_INTERP_TILES:
      tile_make_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
      break;

   case PIXOPS_INTERP_BILINEAR:
      bilinear_make_fast_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
      break;

   case PIXOPS_INTERP_HYPER:
      bilinear_make_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
      break;
   }

   if (filter.n_x == 2 && filter.n_y == 2 &&
       dest_channels == 4 && src_channels == 4 &&
       src_has_alpha && !dest_has_alpha)
   {
#ifdef USE_MMX
      if (found_mmx) {
         line_func = composite_line_22_4a4_mmx_stub;
      } else {
#endif
         line_func = composite_line_22_4a4;
      }
   } else {
      line_func = composite_line;
   }

   pixops_process (dest_buf, render_x0, render_y0, render_x1, render_y1,
                   dest_rowstride, dest_channels, 24, dest_has_alpha,
                   src_buf, src_width, src_height, src_rowstride, src_channels,
                   24, NULL, src_has_alpha, scale_x, scale_y, 0, 0, 0, 0, 0,
                   &filter, line_func, composite_pixel);

   free (filter.weights);
   return PIXOPS_STATUS_SUCCESS;
}


PixopsStatus
pixops_scale (uint8 *dest_buf,
              int render_x0,
              int render_y0,
              int render_x1,
              int render_y1,
              int dest_rowstride,
              int dest_bytes_per_pixel,
              int dest_depth,
              boolean dest_has_alpha,
              const uint8  *src_buf,
              int src_width,
              int src_height,
              int src_rowstride,
              int src_bytes_per_pixel,
              int src_depth,
              uint32 *src_colormap,
              boolean src_has_alpha,
              double scale_x,
              double scale_y,
              PixopsInterpType interp_type)
{
   PixopsFilter filter;
   PixopsLineFunc line_func;

#ifdef USE_MMX
   boolean found_mmx = pixops_have_mmx();
#endif

   if (pixops_support_scale_formats (dest_bytes_per_pixel, dest_depth,
                                     dest_has_alpha, src_bytes_per_pixel,
                                     src_depth, src_colormap, src_has_alpha)
       != PIXOPS_STATUS_SUCCESS)
   {
      return PIXOPS_STATUS_UNSUPPORTED_FORMAT;
   }

   if (scale_x == 0 || scale_y == 0) {
      return PIXOPS_STATUS_FAILURE;
   }

   switch (interp_type) {
   case PIXOPS_INTERP_NEAREST:
      pixops_scale_nearest (dest_buf, render_x0, render_y0, render_x1, render_y1,
                            dest_rowstride, dest_bytes_per_pixel, dest_depth,
                            dest_has_alpha,
                            src_buf, src_width, src_height, src_rowstride,
                            src_bytes_per_pixel, src_depth, src_colormap, src_has_alpha,
                            scale_x, scale_y);
      return PIXOPS_STATUS_SUCCESS;

   case PIXOPS_INTERP_TILES:
      tile_make_weights (&filter, scale_x, scale_y, 1.0);
      break;

   case PIXOPS_INTERP_BILINEAR:
      bilinear_make_fast_weights (&filter, scale_x, scale_y, 1.0);
      break;

   case PIXOPS_INTERP_HYPER:
      bilinear_make_weights (&filter, scale_x, scale_y, 1.0);
      break;
   }

   if (filter.n_x == 2 && filter.n_y == 2 &&
       dest_bytes_per_pixel == 3 && src_bytes_per_pixel == 3) {
#ifdef USE_MMX
      if (found_mmx) {
         line_func = scale_line_22_33_mmx_stub;
      } else {
#endif
         line_func = scale_line_22_33;
      }
   } else {
      line_func = scale_line;
   }

   pixops_process (dest_buf, render_x0, render_y0, render_x1, render_y1,
                   dest_rowstride, dest_bytes_per_pixel, dest_depth, dest_has_alpha,
                   src_buf, src_width, src_height, src_rowstride, src_bytes_per_pixel,
                   src_depth, src_colormap,
                   src_has_alpha, scale_x, scale_y, 0, 0, 0, 0, 0,
                   &filter, line_func, scale_pixel);

   free (filter.weights);
   return PIXOPS_STATUS_SUCCESS;
}


